#include "gnuplot_plugin.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_roots.h>

#define EPSABS 1e-8
#define EPSREL 1e-10
#define MAXTIMES 400
#define INTEG_WORKSPACE_SIZE 1000
#define INTEG_SUBINTERVAL_LIMIT 999
#define DIFF_EPS 1e-5  /*微分用の微小数。小さくしすぎるとこける*/

#define min(x,y) ( (x)<(y) ? (x) : (y))

/* 関数に渡す引数を格納するための構造体 */
struct my_param { double a; double b;};
struct my_param3 { double a1; double a2; double a3;};
struct my_param5 { double a1; double a2; double a3; double a4; double a5;};



/* 超伝導ギャップ計算の際の被積分関数 xがエネルギー、tは規格化温度、dはギャップ */
double gap_integrand(double x, void *p){
  struct my_param *params = (struct my_param *)p;
  double d = (params -> a);
  double t = (params -> b);
  return tanh( sqrt(x*x + d*d)/(2.0*t) )  / sqrt(x*x+d*d) ;
}


/* 超伝導ギャップ計算のためのエネルギー積分を実行する関数 */
double gap_integral(double x, void *p){
  /* xが規格化したギャップ、tが規格化した温度に対応、
     cutoffはcutoffエネルギー（デバイ温度に取る場合が多い）に対応する。 */
  struct my_param *Iparams = (struct my_param *)p;
  double cutoff = (Iparams -> a);
  double t = (Iparams -> b);
  double q = log(2*exp(M_EULER)*cutoff/M_PI);
  double uplim;
  double result;
  double abserr;
  int neval;
  gsl_function G;
  struct my_param Gparams = {x, t};
  gsl_integration_workspace *workspace = gsl_integration_workspace_alloc (INTEG_WORKSPACE_SIZE);
  
  double gap_integrand(double, void *);

  G.function = &gap_integrand;
  G.params = &Gparams;
  
  /* 積分の上限 */
  uplim = cutoff;
  
  gsl_integration_qag(&G, 0.0, uplim, EPSABS, EPSREL, INTEG_SUBINTERVAL_LIMIT, 1, workspace, &result, &abserr);

  gsl_integration_workspace_free (workspace);
  return result - q;
}





/* 超伝導ギャップの計算、tは規格化した温度T/Tc、cutoffはカットオフエネルギー。*/
double gap_main(double t, double cutoff){
  double gap;
  double gap_integral(double, void *);
  int times;
  int status;
  double x_l, x_r;
  
  gsl_function F;
  struct my_param Fparams;

  /* 解を数値的に求める際のWorkspaceの確保  */
  const gsl_root_fsolver_type *T = gsl_root_fsolver_bisection;
  gsl_root_fsolver *workspace_f  = gsl_root_fsolver_alloc (T);
  
  F.function = &gap_integral;
 
  (&Fparams)->a  = cutoff;
  (&Fparams)->b  = t;
  F.params = &Fparams;
  
  gsl_root_fsolver_set (workspace_f, &F, 0.001, 2.0);

  /* 解を求めるループ */
  for(times = 0; times < MAXTIMES; times++)
    {
      status = gsl_root_fsolver_iterate(workspace_f);
      
      x_l = gsl_root_fsolver_x_lower(workspace_f);
      x_r = gsl_root_fsolver_x_upper(workspace_f);
      
      status = gsl_root_test_interval(x_l, x_r, EPSABS, EPSREL);
      /* 集束したらループ停止 */
      if(status != GSL_CONTINUE)
        {
          gap = (x_l+x_r)*0.5;
          break;
        }
    }
  
  /* free */
  gsl_root_fsolver_free(workspace_f);
  
  return gap;
}

 
/* 超伝導体での準粒子エネルギーE(k)を求める関数 kは波数、efはFermi Energy、dはGap */
double ek(double k, double ef, double d){
  return sqrt((k*k-ef)*(k*k-ef) + d*d);
}


/* 比熱計算の際の被積分関数 xが波数*/
double specific_heat_integrand(double x, void *p){
  double ek(double, double, double);
  struct my_param5 *Cparams = (struct my_param5 *)p;
  double e;
  double ef     = (Cparams -> a1);  /* Fermi energy */
  double t      = (Cparams -> a2);  /* 規格化温度 */
  double cutoff = (Cparams -> a3);  /* Cutoff */
  double d      = (Cparams -> a4);  /* gap */
  double d2d    = (Cparams -> a5);  /* gapの2乗の温度微分 */

  e = ek(x, ef, d);   /* 波数xの準粒子エネルギー */
  return x*x*( e*e - t*0.5*d2d ) * ( exp(-e/t)/t ) / ( (exp(-e/t)+1.0) * (exp(-e/t)+1.0) );
}



/* 超伝導状態の比熱を返す関数 tは規格化温度、cutoffはcutoffエネルギー、efはFermi Energy（10000程度） */
double specific_heat_main(double t, double cutoff, double ef){
  double d, d1, dt, d2d;
  double uplim, lowlim, result, abserr;
  gsl_function C;
  struct my_param5 Cparams;
  
  double specific_heat_integrand(double, void *);
  double gap_main(double, double);

  /* 積分するためのWorkspaceの確保 */
  gsl_integration_workspace *workspace = gsl_integration_workspace_alloc (INTEG_WORKSPACE_SIZE); 

  if (t<1.0){
    d = gap_main(t, cutoff);
    dt = DIFF_EPS;
    d1 = gap_main(t+dt, cutoff);
    // gapの2乗の温度微分
    d2d = (d1*d1-d*d)/dt;
  }
  else {
    // Normal state (t>1) ではGapはゼロ
    d = 0.0;
    d2d = 0.0;
  }

  C.function = &specific_heat_integrand;
  (&Cparams)->a1  = ef;
  (&Cparams)->a2  = t;
  (&Cparams)->a3  = cutoff;
  (&Cparams)->a4  = d;
  (&Cparams)->a5  = d2d;
  C.params = &Cparams;

  /* 積分の上限と下限 cutoff または温度の50倍のうちの小さい方 */
  uplim  = sqrt(ef + min(cutoff, t*50) );
  lowlim = sqrt(ef - min(cutoff, t*50) );

  /* 積分の実行 */
  gsl_integration_qag(&C, lowlim, uplim, EPSABS, EPSREL, INTEG_SUBINTERVAL_LIMIT, 1, workspace, &result, &abserr);

  /* Workspaceを開放する */
  gsl_integration_workspace_free (workspace);

  /* 積分結果を温度で割ると比熱になる */
  return result/t;
  
}



/* gnuplotでBCSギャップ関数を使うための関数
   引数は2つで、1つ目が規格化温度、2つ目はカットオフエネルギー（普通の用途では500程度にしておく）*/
DLLEXPORT struct value BCS_gap(int nargs, struct value *arg, void *p){
  
  double gap_main(double, double);
  double t;
  double cutoff;
  struct value r;
  
  /* 引数の数とタイプのチェック */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 2);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[0]);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[1]);

  /* 引数の代入 */
  t = RVAL(arg[0]);
  cutoff = fabs(RVAL(arg[1]));

  /* 結果を返すための構造体のタイプを指定 */
  r.type = CMPLX;

  /* 虚部はゼロ */
  r.v.cmplx_val.imag =   0.0;
  
  /* 実部に結果を代入 */
  if (t >= 1.0){
    /* Normal状態ではギャップゼロ */
    r.v.cmplx_val.real = 0.0;
  }
  else if (t < 0.001){
    /* tがゼロに近いとエラーになるので、tが小さい場合はt=0.001での値で代用 */
    r.v.cmplx_val.real =   gap_main(0.001,cutoff);
  }
  else {
    /* 上記以外ではギャップを計算 */
    r.v.cmplx_val.real =   gap_main(t,cutoff);
  }
  
  return r;
}


/* gnuplotでBCSの比熱(Normal状態の電子比熱係数で規格化)を使うための関数
   引数は3つで、一つ目は規格化温度t、二つ目はcutoff（500程度にする）、三つめはFermi energy（10000程度にする）*/
DLLEXPORT struct value BCS_Cp(int nargs, struct value *arg, void *p){
  
  double specific_heat_main(double, double, double);
  struct value r;
  double t;
  double cutoff;
  double ef;
  
  /* 引数の数とタイプのチェック */
  RETURN_ERROR_IF_WRONG_NARGS(r, nargs, 3);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[0]);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[1]);
  RETURN_ERROR_IF_NONNUMERIC(r, arg[2]);
  
  /* 引数の代入 */
  t  = RVAL(arg[0]);
  cutoff  = fabs(RVAL(arg[1]));
  ef      = fabs(RVAL(arg[2]));

  /* 結果を返すための構造体のタイプを指定 */
  r.type = CMPLX;

  /* 虚部はゼロ */
  r.v.cmplx_val.imag =   0.0;

  /* 実部に結果を代入 */
  if (t >= 1.0){
    /* Normal状態の場合は温度に比例する比熱
      （Normal状態の電子比熱係数で規格化するので、温度そのものを返せばいい*/
    r.v.cmplx_val.real = t;
  }
  else if (t < 0.001){
    /* 温度が十分低い場合は比熱ゼロ い*/
    r.v.cmplx_val.real =   0.0;
  }
  else {
    /* それ以外は比熱を計算しt=1の時の値で割る。*/
    r.v.cmplx_val.real =   specific_heat_main(t,cutoff,ef)/specific_heat_main(1.0,cutoff,ef);
  }

  return r;
}


